/**
 *  The Piece class is the most basic of game Pieces.
 *  All pieces must conform to our physical world.
 */
class Piece{

    /**
     * Create the Piece object.
     * @param {number} width the width of the piece object.
     * @param {number} height the height of the piece object.
     * @param {number} x the position on the canvas in a horizontal sense.
     * @param {number} y the position on the canvas in a vertical sense.
     *
     * @return {Object} the piece.
     */
    constructor(width,height,color,x,y)
    {

        this.width = width;
        this.height = height;
        this.color = color;
        this.x = x;
        this.y = y;
        // physics constants, i.e. what is the gravity in this make believe world?
        this.gravity = 2.0;

        // if something jumps or bounces we need the force that counteracts gravity.
        this.upForce = 0;
        //control the speed.
        this.speedX = 0;
        this.speedY = 0;
        this.gravitySpeed = 0;
        this.type = 0;
        this.frameNo = 0;
        // if a piece is a living piece.
        this.powerUp = false;
        this.crouch = false;
        this.forward = true;
        this.lastActionFrame = 0;
        this.life = 100;

        this.defaults = {};
    }
   /**
    * Extends an object from a new object.
    *
    * 
    */
    configure( settings)
    {
        var extended = this.defaults;

        // for (var key in settings) {
        //     var argument = settings[key];
        //     for (var prop in argument) {
        //         if (Object.prototype.hasOwnProperty.call(argument, prop)) {
        //             extended[prop] = argument[prop];
        //             alert("default: " + this.defaults[prop] + " argument: " + argument[prop]);
        //         }
        //     }
        // }
        
        for (var prop in settings)
        {
            extended[prop] =settings[prop];
        }

        this.defaults = extended;
    }
    /**
     * Set the current frame for the piece in accordance with its world.
     * @param {number} frame the current frame
     */
    setFrame(frame)
    {
        this.frameNo = frame;
    }
    /**
     * A function that returns a boolean to decide whether a piece should render its next state.
     * @param {number} f the frame interval.
     *
     * @return {boolean} true if frame modulus is zero false otherwise. when modules is met then the 
     * current frame is a multiple of the frameinterval.
     */
    canRender(f)
    {

        if ((this.frameNo / f) % 1 == 0) return true;
        return false;

    }
    /**
     * Transforms frames to seconds to allow fine-grain control for next action in the sequence of actions.
     * @param {nubmer} s the second interval that must transpire for the next action.
     *
     * @return {boolean} true if interval has been met false otherwise.
     */
    performAction(s)
    {
        var frameThresh  = ( s /(myGameArea.updateInterval /1000));
        var frameDiff = this.frameNo - this.lastActionFrame;
        if ( frameDiff  >= frameThresh)
        {
 
            this.lastActionFrame = this.frameNo;
            return true;
        }
        return false;

 
    }

    /**
     * Determines whether two objects have collided.
     * @param {Object} other the other object be it a living entity { Bowser, Mario} or an obstacle such as a fireball.
     *
     * @return {boolean} true if object have collided false otherwise.
     */
    collided (other)
    {
        //type 1 are circle objects.
        // all objects excluding Mario are parametric.
        if (this.type == 1)
        {
         
               var distX = Math.abs(this._x - other.x - other.width / 2);
               var distY = Math.abs(this._y - other.y - other.height / 2);

                if (distX > (other.width / 2 + this.radius)) {
                    return false;
                }
                if (distY > (other.height / 2 + this.radius)) {
                    return false;
                }

               
                if (distX <= (other.width / 2)) {
                    return true;
                }

                if (distY <= (other.height / 2)) {
                    return true;
                }


                var dx = distX - other.width / 2;
                var dy = distY - other.height / 2;

                return (dx * dx + dy * dy <= (this.radius * this.radius));

        }
        // rectangular objects { Mario }
        // return the negation of an obvious no collision.
        return  !(
                   ( (this.y + this.height) < other.y) 
                || (this.y > (other.y + other.height)) 
                || ((this.x + this.width) < other.x) 
                || (this.x > (other.x + other.width))
                );
    }
    /**
     * Determines if an object has hit the bottom of the playing scene.
     *
     * @return {boolean} true if the object is at bottom false otherwise.
     */
    isAtBottom()
    {
        
        var floor = frame.height - this.height;
        return this.y >= floor;
    }

    /**
     * Determines if an object is airborne.
     *
     * @return {boolean} true if its vertical position is above 1.75 of its height above the ground.
     */
    isAirBorne()
    {
        var airLevel = frame.height - (this.height * 1.75);
        return this.y < airLevel;
    }

    /**
     * Default action to take against the piece should it hit the bottom.
     * In our world that is to set any velocities and antigravity forces to zero.
     *
     * @return {boolean} returns true if at bottom false otherwise.
     */
    hitBottom()
    {
        var atBottom = this.isAtBottom();

        if (atBottom)
        {
            var floor = frame.height - this.height;
            this.y = floor;
            this.gravitySpeed = 0;
            this.upForce = 0;
        }
        return atBottom;
    }

    /**
     * Default action to take if the piece hits the top of the scene.
     * set antigravity force to zero, allow gravity to slowly take over and cause a descent ( done on update )
     * set the descent speed to fall slightly.
     * @return {boolean} true if hit top false otherwise.
     */
    hitTop(){
        var top = 0;
        if (this.y <= top)
        {
            this.y = top;
            this.gravitySpeed = 0.1; // if you hit the top then give a slight downward acceleration.
            this.upForce = 0;
            return true;
        }
        return false;
    }

    /**
     * Plays and audio file.
     * @param {HTML Audio Object} audio the music to play.
     */
    play(audio)
    {

        if (audio.paused) {
            audio.play();
        }else{

            audio.currentTime = 0
             audio.play();
        }

    }

}