/**
 * This script is the Game Engine.
 *
 * Controls the scene update and protagonists and antagonists actions.
 * Determines when collisions occur and the game state.
 */

//customize the game margins.
var xMargin = 30;
var yMargin = 110;

// game interval speed.
var updateInterval = 16.66;

// the background music during the game.
var backgroundMusic;
// when mario wins.
var levelComplete;
var rDisplay = 10;
var gameOver = false;


// the canvas rect below uses the constants above to establish the play area.
var frame = {};
var context;

/**
 * Convenience function that yields a random value between a range of values.
 * @param  {number} low  the low value in the range.
 * @param  {number} high the high value in the range.
 * @return {number} random variable between low and high.
 */ 
function rand(low,high) {
        return Math.floor(Math.random() * (high ) + low );
}

/**
 * Starts the game
 * @return {void} 
 */
function startGame() 
{
    // set the frame.
    frame = {width: document.body.scrollWidth - xMargin, height: document.body.scrollHeight - yMargin };

    // the pre-game div is where we present the game.
    // at this point we want to remove it from view.
    var pregame = document.getElementById('pre-game');
    pregame.parentNode.removeChild(pregame);
    
    //create mario and bowser and set theme against each other.
    mario = new Player({   
                            // the multiplier at which Mario will walk speedx * walk
                            walk:1.3,
                            // the multiplier at which Mario will run speedx * run
                            run:2.3, 
                            //low jump and highjump.
                            lowJump:7.0, 
                            highJump:12.0, 
                            // regular translation speed.
                            speedX:4.2
                        }  );
    bowser = new Villain({
                            // the speed when bowser walks..
                            walk: 7.0,
                            // the run speed.
                            run:  11.0,
                            // when bumrush...
                            bumrush: 16.0,
                            // the delay between attacks... 
                            attackDelay:0.7,
                            // fireball speed ! should always be faster than bum rush. but its up to you.
                             fireBallSpeed: 25
                        });
    bowser.opponent = mario;
    mario.opponent = bowser;

    // start the game area ie. the canvas drawing etc.
    myGameArea.start();

    // play the game music
    backgroundMusic = new Audio("media/background-music.mp3");
    backgroundMusic.volume = .4;
    backgroundMusic.loop = true;
    backgroundMusic.load();
    backgroundMusic.play();

    // load the level complete audio.
    levelComplete = new Audio("media/levelComplete.mp3");
    levelComplete.volume = 0.4;
    levelComplete.load();


}

/**
 * This variable will hold all game area attributes
 * and initial function.
 * @type {void}
 */
var myGameArea = {

    // our game scene.
    canvas : document.createElement("canvas"),

    /**
     * Starts the Game Area
     * @return {void} nil
     */
    start : function() 
    {
        /*
         Mario is the main game piece, you interface with mario using the keyboard.
         We add an event listener to capture both the keydown and keyup events 
         and pass those to Mario.
         */
        document.addEventListener("keydown", function( e)
        {   
            mario.processUserInput( "keydown" , e ) ; 
        }, false);
        document.addEventListener("keyup", function( e)
        {
               mario.processUserInput( "keyup"  ,e ) ;
                
        }, false);
        
        // set canvas attributes.
        this.canvas.width = frame.width;
        this.canvas.height = frame.height;
        this.context = this.canvas.getContext("2d");

        document.body.insertBefore(this.canvas, document.body.childNodes[document.body.childNodes.length - 1]);
        this.frameNo = 0;

        // game loop * 16.66 is roughly 60 fps.
        this.interval = setInterval(render, updateInterval);
        this.updateInterval = updateInterval;

        // hold both mario and bowser attacks.
        this.bowserAttacks  = [];
        this.marioAttacks   = [];

         // start the background music....
        },

        // on every frame we clear the canvas and redraw.
    clear : function() {
        this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
    }
}

/**
 * Renders the game scene on every frame.
 * @return {void} nil
 */     
function render() 
{ 
    // once a death occurs stop the game music.
   if ( mario.isDead || bowser.isDead)  backgroundMusic.pause();

   /** Game is over once we see a dead mario plunge of the screen. */
    if ( mario.y > frame.height)
            {
        
                clearInterval(myGameArea.interval);
                gameOver = true;
                return;
    }
    /** If bowser dies we wait until he plunges from the screen to start the mario won sequence of events. */
    if( bowser.y > frame.height)
    {
        if(gameOver ) 
        {
                // player wins will be called until the Mario Wins is complete on the screen.
               if ( playerWins() )  {  return;  }
        }
        else{
            gameOver = true;
            levelComplete.play();
            return ;
        }            
    }
    else{
       myGameArea.clear(); 
    }
   
    
   

    //check bowser attacks...
    // check for collision... 
    checkAttackSuccess(myGameArea.marioAttacks, bowser);
    // check for relevance..
    checkAttackStatus(myGameArea.marioAttacks);

    //check bowser attacks...
    // check for collision... 
    checkAttackSuccess(myGameArea.bowserAttacks, mario);
    // check for relevance..
    checkAttackStatus(myGameArea.bowserAttacks);
           

    myGameArea.frameNo += 1;
    
    //update the pieces.
    mario.setFrame(myGameArea.frameNo);
    mario.position();
    mario.update();

    bowser.setFrame(myGameArea.frameNo);
    bowser.attack();
    bowser.position();
    bowser.update();

}

/**
 * Check if an attack was successful i.e. collided with its intended object.                
 * @param  {Piece object array } attacks  an array holding all piece objects.
 * @param  {Piece Object} opponent the one on whom the attacks were directed.
 * @return {void} nil.
 */
function checkAttackSuccess(attacks,opponent)
{
    //loop through attacks.
    for (i = 0; i < attacks.length; i ++) 
    {
        var attack = attacks[i];

        //check for collision.
        if ( attack.collided(opponent))
        { 
            //decrement from the opponent life the strength of the attack.
            // * Bowser attacks are 100%, life sucks for Mario.
             opponent.life -= attack.strength;
             attack.strength = 0;

             // when an attack is successful push it off the screen.
             attack.x = attack.speedX > 0 ? attack.x + frame.width * 2 : attack.x - frame.width * 2;
             // if the current attack destroys the opponent
             // set the life to zero and kill the opponent.
            if ( opponent.life  <= 0 )
            {
                myGameArea.frameNo = 0;
                backgroundMusic.pause();
                opponent.die();
                return;
            }       
        } 
    }
}

/**
 * Check if an attack is obsolete, i.e. out of the screen. When this is so it is removed from memory.               
 * @param  {Piece object array } attacks  an array holding all piece objects.
 * @param  {Piece Object} opponent the one on whom the attacks were directed.
 * @return {void} nil.
 */
function checkAttackStatus(attacks)
{
        // all finished attacks 
         var  finishedAttacks = [];

         for (i = 0; i < attacks.length; i++)
             {
                var attack =  attacks[i];
                attack.position();
                attack.update();
                attack.frameNo = myGameArea.frameNo;

                // get the attack position if it exceeds these bounds add it to the finished attacks array.
                if ( (attack.x  + attack.width * 2 < 0 )  || (attack.x > frame.width + 2 * attack.width))
                {
                        finishedAttacks.push(i);
                }
            }
            // loop through finished attacks and splice the attack array i.e. remove from the array.
            // set the attack object to null.
            for(i = 0 ; i < finishedAttacks.length; i++)
            {
                var j = finishedAttacks[i];
                attack = attacks [j];
                attacks.splice(j,1);
                attack = null;

            }
           
}

/**
 * When the player { mario } wins you want to show a little message.
 * @return {void} nil
 */
function playerWins()
{
    // we want mario wins to start small and animate to 80px.
        if(rDisplay >= 80) {
                    clearInterval(myGameArea.interval);
                    return true;
        }
        myGameArea.clear();
        var ctx = myGameArea.context;
        var font = rDisplay+"px Arial";


        ctx.fillStyle = "#fff";
        ctx.strokeStyle = "red";
        ctx.lineWidth = "5px";

        ctx.font= font;

        ctx.fillText("Mario Wins!!",frame.width/3 , frame.height/3);
        ctx.strokeText("Mario Wins!!",frame.width/3 , frame.height/3);
        rDisplay+= 1;

        return false;
    
}
/**
 * calculate if the interval is a multiple of n
 * @param  {number} n the frame interval
 * @return {boolean}   true if modulus is 0, false otherwise.
 */
function everyinterval(n) {
    if ((myGameArea.frameNo / n) % 1 == 0) {return true;}
    return false;
}
